import sys
import argparse
import numpy as np
import subprocess as sp

from twigen import twi
from twigen import gen
from twigen import gui

def main():
  tcd = sp.Popen(['tcd', '--client-id', 'x024uk066qafsuz8bqz5mp5jsiio7s'], stdout=sp.DEVNULL, stderr=sp.DEVNULL)
  parser = argparse.ArgumentParser(description="Twitch VOD editor/downloader and highlights generator.")
  parser.add_argument("-u", "--urls", help="URLs or IDs of the VODs to process")
  parser.add_argument("-q", "--quiet", action='store_true', help="do not print what is being done")
  parser.add_argument("-s", "--sample", type=int, default=60, help="duration of each sample in seconds, default=60")
  parser.add_argument("-d", "--delay", type=int, default=30, help="start recording DELAY seconds early in each sample, default=30")
  parser.add_argument("-g", "--group", action='store_true', help="process a list of urls into a single output file")
  parser.add_argument("-l", "--check", action='store_true', help="print how long the output file would be with this options")
  parser.add_argument("-T", "--tmp", action='store_true', help="do not save downloaded data to current directory")
  parser.add_argument("-M", "--muted", action='store_true', help="include muted portions of the VOD")
  parser.add_argument("-i", "--include", help="include the specified samples")
  parser.add_argument("-Z", "--signal", action='store_true', help="look for !twigen in chat messages and create a sample of that moment")
  outg = parser.add_mutually_exclusive_group()
  outg.add_argument("-o", "--output", help="name of the output file")
  outg.add_argument("-S", "--segs", action='store_true', help="do not concatenate samples")
  chatg = parser.add_mutually_exclusive_group()
  chatg.add_argument("-C", "--nochat", action='store_true', help="do not download chat data, process included samples only")
  chatg.add_argument("-m", "--min", type=int, help="minimum length of the output file when generating highlights")
  chatg.add_argument("-y", "--hype", type=float, default=0.6, help="float number between [0,1] used by the highlights generator for selecting samples, default=0.6")
  args = parser.parse_args()

  if len(sys.argv) < 2:
    gui.run()
    return

  urls = args_require(args)

  group, group_check = [], []
  for u in urls:
    if not args.quiet:
      print("Starting new process ... ", end='')
    url, vid = twi.read_url(u)
    proc = twi.Proc(vid=vid, tmp=args.tmp)
    link = twi.link(url)
    m3u8 = twi.m3u8(proc.outdir, link)
    total_len, ts_len = twi.parse_m3u8(m3u8)
    ts = twi.parse_ts(m3u8)
    tslink = twi.tslink(link)
    sample_len = gen.sl(args.sample, ts_len)
    if not args.quiet:
      print("Done.")

    if not args.output:
      out = "twigen_"+vid+".mp4"
    else:
      if len(urls) > 1 and not args.group:
        parser.error("--output not allowed with multiple urls, unless --group is used")
      out = args.output

    if args.nochat:
      if not args.include:
        parser.error("--include requiered with --nochat")
      else:
        ts = twi.uts(gen.includes([], args.include, total_len), ts, ts_len)
        ts = twi.dvod(proc.tsdir, ts, tslink, muted=args.muted, quiet=args.quiet)
        twi.con(proc.tsdir, ts, out, quiet=args.quiet)
        return

    chat = twi.dchat(proc.outdir, vid, quiet=args.quiet)
    data, sigs = gen.parsec(chat, sample_len, total_len, signal=args.signal)
    data = gen.norm(data)

    if args.min:
      if not args.quiet:
        print("Adjusting output length ... ", end='')
      ts = argmin(data, ts, ts_len, mins=args.min*60, t_len=total_len, inc=args.include, sigs=sigs, delay=args.delay)
      if ts is None:
        return
      if not args.quiet:
        print("Done.")
    elif args.hype:
      data = gen.do_delays(gen.cut(data, args.hype), gen.delays, args.delay)
      ts = twi.uts(gen.includes(gen.samples(data), args.include, total_len, sigs), ts, ts_len)

    if args.check:
      if args.hype == 0:
        if args.group:
          group_check.append(str(total_length))
          continue
        else:
          print("Length of output file with current settings: "+str(total_len))
          return
      if args.group:
        group_check.append(str(twi.conlen(ts, ts_len)))
        continue
      else:
        print("Length of output file with current settings: "+str(twi.conlen(ts, ts_len)))
        return

    if args.hype == 0:
      ts = twi.dvod(proc.tsdir, ts, tslink, muted=args.muted, parse_ts=twi.parse_ts, quiet=args.quiet)
    else:
      ts = twi.dvod(proc.tsdir, ts, tslink, muted=args.muted, quiet=args.quiet)

    if args.group:
      tspaths = []
      for t in ts:
        tspaths.append(proc.tsdir+t)
      group.extend(tspaths)
    else:
      if not args.segs:
        twi.con(proc.tsdir, ts, out, quiet=args.quiet)
      else:
        twi.conpart(proc.tsdir, ts, out, sample_len, ts_len, quiet=args.quiet)

  if args.group:
    if args.check:
      checks = ', '.join(group_check)
      print("Length of output file with current settings: "+checks)
      return
    if not args.segs:
      twi.con(proc.tsdir, group, out, quiet=args.quiet, group=True)
    else:
      twi.conpart(proc.tsdir, group, out, sample_len, ts_len, quiet=args.quiet, group=True)

# additional args checks
def args_require(args):
  if not args.urls:
    parser.error("--urls requiered")
  urls = args.urls.replace(' ', '').split(',')
  if args.nochat and args.signal:
    parser.error("--nochat not allowed with --signal")
  if args.group:
    if not args.output:
      parser.error("--output requiered with --group")
    if len(urls) < 2:
      parser.error("more than one url or video id requiered with --group")
  return urls

# adjust lenght of output according to --min
def argmin(data, ts, ts_len, mins, t_len, delay, sigs, inc=None): #TODO refactor
  x, e, b = 0.5, 1, 0
  check, plen = (mins + 60), 0
  safe, lock = 0, 60 #iterations control
  odata, ots = np.copy(data), ts
  while True:
    data = gen.do_delays(gen.cut(data, x), gen.delays, delay)
    ts   = twi.uts(gen.includes(gen.samples(data), inc, t_len, sigs=sigs), ots, ts_len)
    last = plen
    plen = twi.conlen(ts, ts_len).total_seconds()
    if plen < mins:
      e = x
      x -= (e-b)/2
    elif plen > check:
      b = x
      x += (e-b)/2
    else:
      return ts
    if plen == last and plen > mins:
      return ts
    data = odata
    safe += 1
    if x > 1 or x < 0 or safe == lock:
      print("--min failed. x: "+str(x)+" safe: "+str(safe))
      return None
